(in-package "AP5")

#|
RelAb specifications often use unary relations to implement concepts that, in
an object-oriented language (CLOS, C++) we would expect to find implemented
as classes.  (other unary relations are not OO class-like at all -- e.g.,
they correspond to global sets or boolean-valued functions.) I see two 
rational reasons why programmers would be tempted to  use unary relations to
implement classes with OO-class-like semantics, rather than using OO
language interfaces (e.g., CLOS) directly:

(1) ability to enumerate (generate) instances. This is not really inherent in
    unary relations -- it depends on the implementation chosen.  The AP5
    library contains severaly implementations that provide generability.
    The "basetype" implementation, in particular, supports generability within
    a class DAG.  Other implementations can support generability more
    efficiently, but do not support a DAG.  Use of an
    enumerable unary relation permits a class restriction to appear as part of 
    a logic formula, WITHOUT commitment as to whether that restriction is to
    be used a a GENERATOR of candidate objects or simply as a FILTER.  

(2) transaction semantics -- for an ATOMIC unary relation, new objects do not
    appear as members of the relation until the transaction is committed,
    and their creation can be used as a clause in active db monitors.
    OO classes do not provide anything analogous to transaction-mediated
    creation.

OO languages (e.g. CLOS), on the other hand, offer advantages over simple
unary relations:
(1) methods can be specialized on the basis of OO types, but
    not on the basis of unary relation membership.
(2) there is a well thought-out initialization protocol for OO objects.
    RelAb does not address object creation at all, leaving
    it to the programmer to implement initialization for each type
    in terms of some combination of procedures and consistency rules.
    In particular, ASSERTION of a unary relation is in no way
    tied to object ALLOCATION.  (AP5's CREATE macro provides a very superficial
    connection, but is not useful for evolutionary prototyping because
    there is no annotation-based means of controlling the implementation of
    objects allocated via this macro.)


It would be nice to have the advantages of both worlds.  Fortunately,
RelAb unary relations impose no inherent restriction on the implementation
of the objects which populate them, or the algorithms used to test and
generate them.   For a relational Lisp, e.g., we should provide an easy way
to define and pair a CLOS class C with a unary relation that corresponds to
the set of instances of all classes below C in the CLOS DAG.  The 
relation should supports membership testing and either  or both of:
(A) enumerability
(B) transaction semantics

(There doesn't seem to be much to gain by factoring out membership testing,
at least in the case of relational lisp, because CLOS supplies that for
all its classes whether or not a program makes use of it.  I believe this
is also true for the current C++ standard, althought there may be optimizations
for programs that don't require it.)

           Programmer Interfaces (Relational Lisp orientation)

A declaration macro, DEFINE-CLOS-RELATION, with all the pieces of DEFCLASS, 
which defines a unary relation in addition to the class.  The macro allows
the programmer to specify whether the class is to support enumerability,
transactions, or both.  [It is no doubt desirable to also provide
new slot options that implicitly define binary relations
implemented as slots, but that is quite independent of this proposal and
far easier to implement.]

New instances should be created by the usual OO syntax --
MAKE-INSTANCE in the case of relational lisp, constructors in the case
of relC++.  When executed inside a transaction, the new instance is
not visible as a member of the relation until the transaction is
committed.  It is only visible in the "PROPOSED" temporal relation
until committed.

C++ provides interfaces for the programmer to express when an object should
be "released".  In the case of an enumerable class, the object should
no longer appear in the enumeration after this.  Lisp does not provide
such an interface, relying on garbage collection as its only means of
reclaiming objects.  At least in the case of ENUMERABLE classes,
the Lisp programmer needs a way to explicitly remove an object
from the class.  In the case of enumerable transaction semantics classes,
the AP5 "undo" operation also necessitates some means of removing the
object from its class.  For relational lisp, the programmer interface
for this operation will be (-- C x), where C is the name of the
class/unary relation.  [NOTE: the oddly-behaved "destroy" operation that 
appears in some prototype AP5 programs is not impacted by this at all.]

My lisp implementation of transaction-semantics OO relations
actually defines two clos classes for each OO relation.
One, which has the programmer-supplied class name C, has as its direct
instances those objects that have been created by make-instance but not
yet committed (or committed but later freed by "--").  The second,
a subclass of C named COMMITTED-C, holds the instances created
in transactions that have been successfully committed.  [MAKE-INSTANCE
behaves as its own transaction if it is not executed within one.]
For this implementation, at least, the programmer can use the name
COMMITTED-C as a specializer for methods that are only to apply to
"real", rather than "proposed", Cs.  But I'm not sure this is of any
value.  It is a byproduct of,  not the motivation for, the implementation.

            Mixing Regular classes and Relation Classes
I haven't thought it through completely.  At least the following useful mixture
will work:  A class defined by DEFINE-CLOS-RELATION may inherit from a class M
that is a "pure mixin", by which I mean that neither M nor any superclass of
M is ever directly instantiated.


--------------------------------------------------
                     Implementation Plan

For the time being, assume that the entire DAG of such classes below
any one are identical in which capabilities they support.  That is, if
S2 inherits from S1, then either both of them support enumeration or
neither supports it, and either both have transaction semantics or
neither does.  I suppose it does make sense for C-SUB to be enumerable
even if C is not.

I now discuss an implementation strategy for each of the three cases.
In the following, there is a 1-1 correspondence between CLOS classes
and unary relations -- I think it will be obvious when I am talking
about the relation and when I am talking about the class.

 ENUMERABLE classes WITHOUT transaction semantics:

  We provide an INSTANCES slot on the class object, holding
  a list of the current direct instances.  The last action of
  MAKE-INSTANCE for class E is to add the new instance to the list in E's
  INSTANCES slot.  To enumerate E, we enumerate the objects in E's INSTANCES
  slot, and, recursively, the instances of all direct subclasses of E.
  CLOS always maintains a list of direct subclasses on each class.
  The generator for E must traverse the class DAG below E, enumerating each  
  class only once.  A CLOS object has only one direct type, and will appear
  only in the instances list for that type, so the generator need not worry 
  about encountering the same instance multiple times.  If we knew that the
  DAG was a tree, as it commonly is, we would not need to worry about
  reencountering the same class, but CLOS does not keep track of which
  classes are the roots of pure trees.  (I think that would not be hard to
  do with MOP interfaces, but the cases where the gain is significant may
  be uncommon.)  EVERY class in the DAG below the class being enumerated
  must supply a method for INSTANCES.  This requirement could be enforced or
  not -- it is a requirement on the class hierarchy, not the run-time behavior.

  Membership testing is simply via TYPEP, which already incorporates the
  correct inheritance semantics.

  There is no ADDER or DELETER.

  CHANGE-CLASS will work to move objects between two classes of category A,
  or between a category A class and a non-relational class.  In particular,
  doing (change-class x 'standard-object) will remove the object from
  all enumerations (and conceivably make it gc-able).


      classes with TRANSACTION semantics:
If the programmer defines an atomic clos class relation named FOO, two
ap5 unary relations are defined: FOO-DIRECT (implementation
INDIVIDUAL) and FOO (derivation INDIVIDUAL-DERIVED).  The programmer
is not expected to know about FOO-DIRECT.  There are also two clos
classes defined: Foo and Committed-Foo.  Committed-Foo inherits from
Foo, but has no additional slots.  [The actual symbol names are FOO
and COMMITTED-FOO -- I will use mixed case when naming classes and
upper case for relations.]  The programmer is not expected to know
about Committed-Foo.  When delta is NIL, FOO-DIRECT(x) is true iff 
(class-of x) is the class Committed-Foo.

FOO(x) is true iff FOO-DIRECT(x) is true or FOO_SUB(x) is true for
some relation FOO-SUB corresponding to a subclass Foo-Sub of Foo.
Class inheritance is defined in the define-clos-relation macro, which
mirrors the syntax of DEFCLASS.

It is easiest to think of the class Committed-Foo as corresponding to 
               (previously (FOO x))
and the class Foo as corresponding to a derived relation 
        ((x) s.t. (or (previously (FOO x)) (proposed (FOO x))))


[Alternatively. you could make an analogy to AP5's "bastypes" --
 If FOO were a basetype, then FOO-DIRECT would be
       ((x) s.t. (classification x FOO))
]



A MAKE-INSTANCE method applicable to Foo does
(++ FOO-DIRECT x) for the fully initialized object x.  The ADDER for
FOO-DIRECT uses change-class to move x from class Foo to subclass
Committed-Foo.  If FOO was declared enumerable (the default), then
UPDATE-INSTANCE-FOR-DIFFERENT-CLASS (always called by change class in CLOS)
adds x to a maintained list of instances in a slot of the Committed-Foo
class object.

The DELETER for FOO-DIRECT removes an object from the maintained list
changes the class of x from Committed-Foo to Foo.
If FOO is enumerable, UPDATE-INSTANCE-FOR-DIFFERENT-CLASS takes care
of removing  the object from the maintained list of instances.


The generator for FOO (if it is enumerable)  recurses through the 
clos hierarchy.  If FOO is atomic, delta is taken into account.
This is done using MAP-UPDATES -- the testing/generating code is
shared by all clos relation types.  The same is true for FOO's Tester,
if delta is non-nil.  But if delta is nil, the tester is simply 
           (TYPEP x Committed-Foo)

If FOO is atomic, then the following triggers are asserted:
  (FOO-DIRECT t CLOS-ATOMIC-TRIGGER-DIRECT FOO t)
  (FOO-DIRECT nil CLOS-ATOMIC-TRIGGER-DIRECT FOO nil)

If FOO and a subclass FOO-SUB are atomic, then the following triggers are 
asserted:
      (FOO-SUB t CLOS-ATOMIC-TRIGGER-SUPER FOO t)
      (FOO-SUB nil CLOS-ATOMIC-TRIGGER-SUPER FOO nil)

The triggering functions are shared by all clos relation classes.

FOO has a DELETER which does (-- C_DIRECT x) for whatever class C x
belongs to.  (AP5 ensures that the deleter will not be executed
if x does not belong to FOO or one of its subclasses.  If C is enumerable,
the effect of C's deleter, described above, removes x from the list of
instances, so it may be GCable.)  There is no ADDER for FOO.

If FOO-SUB is declared to inherit from FOO, then both must have the
same atomicity and generability declarations.  (i.e., either both atomic
or neither atomic, either both enumerable or neither enumerable.)
I suppose it does make sense for FOO-SUB to be enumerable even if FOO
is not, but currently an error is signalled (i think).

DEFINE-CLOS-RELATION complains if you declare the relation to be both nonatomic
and non-enumerable, because there don't seem to be any benefits of declaring
a relation in that case, since ap5 already lets you use type names as
unary relations for testing purposes.


Although I currently do declare a generator for FOO-DIRECT, I don't think
it is actually used for anything, and need not be declared.  Conceivably an
argument could be made for "advertising" this relation to the programmer.


---------- cop out      ---------------------

"It is an error" for the programmer to use CHANGE-CLASS explicitly
to move objects into or out of the FOO and COMMITTED-FOO classes.
(It is ok, however, to use change class to move objects in and out of
enumerable, nonatomic clos relation classes.)

"It is an error" for the programmer to define classes (other than through
the define-clos-relation macro) that inherit from either Foo or Comitted-Foo.

A programmer should not supercede the make-instance method for FOO. 
Augmentation is ok.

There is undoubtedly more that really needs to be done to get better
cost/size numbers -- I have punted.

------------ programming note --------------
       (++ R (make-instance 'Foo) x y z) 
is NOT an atomic operation -- AP5 translates the ++ macro such that
the tuple element expressions are computed OUTSIDE an atomic update to
R.  So if you want atomicity, be sure to write
    (atomic (++ R (make-instance 'Foo) x y z))
or use the CREATE macro, which also provides atomicity.



|#


;; by making ENUMERABLE-CLASS-MIXIN an ancestor of the metaclass of a class C,
;; C wlll record its direct instances in a slot.  The recording will take
;; place at the end of make-instance, unless the dont-record keyword is
;; provided in the call to make-instance.

(defclass enumerable-class-mixin ()
     ((instances :initform () :accessor instances))
     (:documentation  "Lets a class record its instances."))

(defmethod make-instance ((class enumerable-class-mixin) &rest initargs
                          &key ((dont-record dont-record) nil)
			  &allow-other-keys)
  #-allegro(declare (ignore initargs))
  (let ((instance (call-next-method)))
    (unless dont-record
      (when (typep instance class)(push instance (instances class))))
    instance))

;; enumerable-class is a metaclass to use for clos classes 
;; 
(defclass enumerable-nonatomic  (enumerable-class-mixin  standard-class) ()
 (:documentation "This is the metaclass for implementing nonatomic, enumerable types."))


(defclass atomic-class-potential-mixin ()
  ((direct-rel :accessor direct-rel)
   (committed-subclass :type class :accessor committed-subclass))
  (:documentation  "A mixin for the metaclass used for the \"potential\" class
of an ATOMIC class."))

(defclass atomic-class-committed-mixin ()
  ((decommitted-superclass :type class :accessor decommitted-superclass)
   (clos-relation :accessor clos-relation
		  :documentation "stored for more efficient triggering"))
  (:documentation  "A mixin for the metaclass used for the \"committed\" class
of an ATOMIC class."))

(defmethod direct-rel ((c atomic-class-committed-mixin))
  (direct-rel (decommitted-superclass c)))


(defmethod make-instance ((class atomic-class-potential-mixin) &rest initargs)
  (checkupdateallowed class t) ;; AP5 has a bad interface here.
  (let ((obj (apply #'call-next-method class
		    #+lucid :allow-other-keys #+lucid t
		    ;;lucid bug -- doesn't seem to completely handle keyword 
                    ;;args to make-instance that are not in the KEYWORD pkg
		    'dont-record t initargs)))
    (++ funcall (direct-rel class) obj)
    obj))

(declaim (inline  commit decommit))
(defun commit (x) (commit-using-class x (class-of x)))
(defun decommit (x) (decommit-using-class x (class-of x)))

(defmethod commit-using-class (instance (mc atomic-class-potential-mixin))
  (change-class instance (committed-subclass mc)))

(defmethod decommit-using-class (instance (mc atomic-class-committed-mixin))
  (change-class instance (decommitted-superclass mc)))

(defclass atomic-potential
    (atomic-class-potential-mixin standard-class) ()
	  (:documentation "The metaclass for the \"potential\" class of the
class pair that provides transaction semantics."))

(defclass nonenumerable-atomic-committed
    (atomic-class-committed-mixin standard-class) ()
    (:documentation "The metaclass for the \"committed\" class of the
class pair that provides transaction semantics without enuerability."))

(defclass enumerable-atomic-committed
    (atomic-class-committed-mixin enumerable-class-mixin standard-class) ()
   ;; ORDER OF ABOVE IS IMPORTANT - ATOMIC BEFORE ENUMERABLE!!
    (:documentation "The metaclass for the \"committed\" class of the
class pair that provides transaction semantics with enuerability."))



(defgeneric committed-class (x))
(defmethod committed-class ((c class)) nil)
(defmethod committed-class ((c atomic-class-committed-mixin)) t)
(defun committed (x) (committed-class (class-of x)))


#| ec-mixin-object is a common superclass for any class
   whose metaclass inherits from ENUMERABLE-CLASS-MIXIN
   this class is needed so that methods for 
   UPDATE-INSTANCE-FOR-DIFFERENT-CLASS can be defined to
   maintain the contents of the INSTANCES slot.  |#

(defclass ec-mixin-object () ())


(let ((emoclass (find-class 'ec-mixin-object)))
#-lucid ;; is this standard CL now?
 ;;; ensure that all classes whose metaclass includes ENUMERABLE-CLASS-MIXIN
 ;;; will inherit from ec-mixin-object
 (defmethod shared-initialize :around
	   ((x ENUMERABLE-CLASS-MIXIN) slot-names &rest initargs
	    &key direct-superclasses &allow-other-keys)
  (if (some #'(lambda (x) (subtypep x emoclass)) direct-superclasses)
      (call-next-method)
    (apply #'call-next-method x slot-names
	     :direct-superclasses (append direct-superclasses (list emoclass))
	     initargs)))

#+lucid
(lcl:defadvice (clos-sys:common-add-named-class ec-mixin-object)
    (proto name superclasses slots options)
  (if (typep proto 'enumerable-class-mixin)
     (lcl:advice-continue proto name
	    (if (some #'(lambda (x) (subtypep x emoclass)) superclasses)
		superclasses
	      (append superclasses (list emoclass)))
	    slots options)
    (lcl:advice-continue proto name superclasses slots options)))
)

#| acp-mixin-object is a common superclass for any class
   whose metaclass inherits from ATOMIC-CLASS-POTENTIAL-MIXIN
   this class is needed so that the DONT-RECORD keyword
   will be allowed in the initialize-instance calls. |#


(defclass acp-mixin-object () ())
(defmethod initialize-instance :after
	   ((c acp-mixin-object) &key ((dont-record dont-record) nil)
				      &allow-other-keys)
  (declare (ignore dont-record))
  nil)

(let ((apmclass (find-class 'acp-mixin-object)))
#-lucid ;; is this standard CL now?
 (defmethod shared-initialize :around
	   ((x ATOMIC-CLASS-POTENTIAL-MIXIN) slot-names &rest initargs
	    &key direct-superclasses &allow-other-keys)
  (if (some #'(lambda (x) (subtypep x apmclass)) direct-superclasses)
      (call-next-method)
    (apply #'call-next-method x slot-names
	     :direct-superclasses (append direct-superclasses (list apmclass))
	     initargs)))
)


;; These methods for update-instance-for-different-class make sense
;; for non-atomic enumerable classes.  I have not thought out
;; the matter for atomic enumerable classes.
(defmethod update-instance-for-different-class :after
     ((previous t) (current ec-mixin-object) &rest initargs)
  (declare (ignore initargs))
  (let ((newclass (class-of current)))
    (unless (eq (class-of previous) newclass)
      (push current (instances newclass)))))

(defmethod update-instance-for-different-class :after
	   ((previous ec-mixin-object) (current t) &rest initargs)
  (declare (ignore initargs))
  (let ((oldclass (class-of previous)))
    (unless (eq (class-of current) oldclass)
      (setf (instances oldclass)
	(delete current (instances oldclass) :test #'eq :count 1)))))


;; FOR NOW, it is an error FOR A PROGRAMMER TO USE CHANGE-CLASS DIRECTLY
;; ON INSTANCES OF EITHER THE POTENTIAL OR COMMITTED CLASSES.



(defmethod #+lucid clos-system:validate-superclass
	   #+allegro clos:validate-superclass
	   #-(or lucid allegro) validate-superclass
	   ((sub enumerable-nonatomic)
	    (sup standard-class))
  t)

(defmethod #+lucid clos-system:validate-superclass
	   #+allegro clos:validate-superclass
	   #-(or lucid allegro) validate-superclass
	   ((sub standard-class)
	    (sup enumerable-class-mixin))
   ;; all subclasses of an enumerable class must also be enumerable
	   nil)

(defmethod #+lucid clos-system:validate-superclass
	   #+allegro clos:validate-superclass
	   #-(or lucid allegro) validate-superclass
	   ((sub atomic-class-potential-mixin)
	    (sup standard-class))
   t)

(defmethod #+lucid clos-system:validate-superclass
	   #+allegro clos:validate-superclass
	   #-(or lucid allegro) validate-superclass
	   ((sub atomic-class-committed-mixin)
	    (sup atomic-class-potential-mixin))
	   t)

(defmethod #+lucid clos-system:validate-superclass
	   #+allegro clos:validate-superclass
	   #-(or lucid allegro) validate-superclass
	   ((sub atomic-class-committed-mixin)
	    (sup atomic-class-committed-mixin))
	   t)

(defmethod #+lucid clos-system:validate-superclass
	   #+allegro clos:validate-superclass
	   #-(or lucid allegro) validate-superclass
	   ((sub atomic-class-committed-mixin)
	    (sup (eql (find-class 'ec-mixin-object))))
   t)

(defmethod #+lucid clos-system:validate-superclass
	   #+allegro clos:validate-superclass
	   #-(or lucid allegro) validate-superclass
	   ((sub atomic-class-committed-mixin)
	    (sup standard-class))
 t)

(defmethod #+lucid clos-system:validate-superclass
	   #+allegro clos:validate-superclass
	   #-(or lucid allegro) validate-superclass
	   ((sub standard-class)
	    (sup atomic-class-potential-mixin))
	   nil)

(defmethod #+lucid clos-system:validate-superclass
	   #+allegro clos:validate-superclass
	   #-(or lucid allegro) validate-superclass
	   ((sub standard-class)
	    (sup atomic-class-committed-mixin))
	   nil)

(defrelation clos-class-for-relation
    :types (relation class)
    :type-enforcements (:default :none)
    :representation treeprop 
      ;; should use a better representation, with unhashed access in both
      ;; directions, not just from relation to class.
      ;; There is no requirement for total generability.
    )

(let ((ecm (find-class 'enumerable-class-mixin)))
 (defun clos-enumerable-generator (R)
  ;; an efficient ap5 generator for all instances of all classes of the
  ;; CLOS DAG rooted at the class corresponding to relation R
  ;; This if for nonatomic relations only.
  (let ((typesremaining (list (clos-class-for-relation R ?)))
	(typesgenerated ())
	(instances ()))
    #'(lambda ()
	(loop
	  (when instances (return (values nil (pop instances))))
	  (unless typesremaining (return t))
	  ;; move first entry from remaining into generated, 
	  ;; this is equivalent to (push (pop typesremaining) typesgenerated)
	  ;; but w/o consing
	  (let ((head typesremaining))
	    (setf typesremaining (rest typesremaining))
	    (setf (rest head) typesgenerated
		  typesgenerated head))
	  (setf instances
	    (when (typep (first typesgenerated) ecm)
		(instances (first typesgenerated))))
	  (dolist (tp (#+(or lucid allegro)clos:class-direct-subclasses
		       #-(or lucid allegro)class-direct-subclasses
			 (first typesgenerated)))
	    (unless (member tp typesgenerated)
	      (pushnew tp typesremaining))))))))


(defun clos-enumerable-nonatomic (name keys class-name)
 (warn-ignored-keys name keys (:arity  :tester :generator :equivs)
  `(:representation individual :arity 1 :equivs (eql)
    :nonatomic t
    :tester (lambda (&rest ignore) (declare (ignore ignore))
		    '(lambda (ignore a1)
		      (declare (ignore ignore))
		      (typep a1 ',class-name)))
    :generator (lambda (subsetflg vars rel &rest ignore)
		 (declare (ignore subsetflg vars ignore))
		 `((template (output))
		   (effort ,(* 3 (relationsize '(x) (list rel 'x))))
		   (initialstate clos-enumerable-generator)))
    :deleter  (lambda (&rest ignore) (declare (ignore ignore))
		 '(lambda (relation object) (declare (ignore relation))
		   (change-class object 'standard-object)))
    :alsofns ((lambda (rel) (setf (clos-class-for-relation rel)
			      (find-class ',class-name)))))))

(defun direct-rel-name (class-name)
  (intern (format nil "~a_DIRECT" class-name)(symbol-package class-name)))


(defun del-from-clostype (relation object)
  ;; should only  be called as the DELETER of an atomic clos class relation
  ;; object is currently an instance of some class within the hierarchy
  ;; below the class corresponding to RELATION
  (insist "supposed to delete it" not (funcall relation object))
  (-- funcall (direct-rel (class-of object)) object))

(labels ((ceasing (object class)
	      ;; since we are not supporting change-class BETWEEN clos
	      ;; relation classes, the only thing to consider is
	      ;; the removal of object from its immediate class
	   (let ((dirrel (direct-rel (decommitted-superclass class))))
	     (map-updates #'(lambda (rel tpl tv)
			      (declare (ignore rel tv))
			      (when (eql (first tpl)  object)
				(return-from ceasing t)))
			  :rel dirrel :tv '(nil))
	     nil))
	 (becoming (object class)
	    ;; the only supported way for an object to ENTER an atomic
	    ;; clos relation class is to move from POTENTIAL to COMMITTED
	   (let* ((prevclass (class-of object))
		  (dirrel (direct-rel prevclass)))
	     (when (eq prevclass (decommitted-superclass class))
	       (map-updates #'(lambda (rel tpl tv)
				(declare (ignore rel tv))
				(when (eql (first tpl)  object)
				  (return-from becoming t)))
			    :rel dirrel :tv '(t))
	       nil))))
	   
  (defun clos-atomic-direct-tester (object class)
    ;; class is the commited-class
    (let* ((prevclass (class-of object))
	   (prevstat (eql prevclass class)))
      (if delta
	  (if prevstat
	      (not (ceasing object class))
	    (becoming object class))
	prevstat)))

  (defun clos-atomic-tester (object class)
    ;; class is the commited-class
    (let ((prevstat (typep object class)))
      (if delta
	  (let ((prevclass (class-of object))) ;; may be a subclass of class
	    (if prevstat
		(not (ceasing object prevclass))
	      (and (typep prevclass 'atomic-class-potential-mixin)
	       (becoming object (committed-subclass prevclass))))) 
	prevstat)))
)

(defun clos-atomic-trigger-direct (derivedrel adds dels)
  ;; This trigger is responsible for telling AP5 that R_DIRECT(x) leads
  ;; to R(x)  (both add and delete)
  (map-updates #'(lambda (rel tpl tv) (declare (ignore rel))
		   (if tv
		       (++ apply derivedrel tpl)
		     (-- apply derivedrel tpl)))
	       :rel (direct-rel (clos-class-for-relation derivedrel))
	       :tv (if adds
		       (if dels '(t nil) '(t))
		     '(nil))))

(defun clos-nonenumerable-atomic (name keys class-name comclass-name)
 ;; comclass-name is the name of the COMMITTED class
 (warn-ignored-keys name keys (:tester :definition :arity)
  `(:arity 1
    :tester (lambda (&rest ignore) (declare (ignore ignore))
	      '(lambda (ignore a1) (declare (ignore ignore))
		 (clos-atomic-tester a1 (load-memo (find-class ',comclass-name)))))
    :testeffort 10
    :derivation individual-derived
    :trigger ((,(direct-rel-name class-name) + clos-atomic-trigger-direct +)
	      (,(direct-rel-name class-name) - clos-atomic-trigger-direct -))
    :deleter (lambda (&rest ignore) (declare (ignore ignore))
               'del-from-clostype)
    :alsofns ((lambda (rel)
		(setf (clos-class-for-relation rel)
		  (find-class ',comclass-name)
		  (clos-relation (find-class ',comclass-name))
		  rel))))))

(defun clos-enumerable-atomic-generator (R)
  ;; an efficient ap5 generator for all instances of all classes of the
  ;; CLOS DAG rooted at the class corresponding to relation R
  (if delta
      (let ((typesremaining (list (clos-class-for-relation R ?)))
	    (typesgenerated ())
	    (instances ())
	    (class_declass ()))

	;; (A) when instances /= (), produce members of
	;;     instances which are NOT also members of class_declass.
	;; (B) when instances = () and class_declass /= (), produce members of
	;;     class_declass
	;; (C) when instances = () and class_declass = (), it is time to move 
	;;     to next class in typesremaining.

	(flet ((setup_becoming (class)
		 (setf class_declass ())
		 (map-updates
		  #'(lambda (rel tpl tv) (declare (ignore rel tv))
			(push (first tpl) class_declass))
		  :rel (direct-rel (decommitted-superclass class))
		  :tv '(t)))
	       (setup_ceasing (class)
		 (setf class_declass ())
		 (map-updates
		  #'(lambda (rel tpl tv) (declare (ignore rel tv))
			(push (first tpl) class_declass))
		  :rel (direct-rel (decommitted-superclass class))
		  :tv '(nil))))
	#'(lambda ()
   	    (loop named typeloop
	      do (cond
		  ((and class_declass instances)
		   (loop while instances
		       do (let* ((i (pop instances))
				 (valid (member i class_declass)))
			    (unless instances
			      (setup_becoming (first typesgenerated)))
			    (when valid
			      (return-from typeloop (values nil i))))))
		  (instances
		   (return-from typeloop
		     (multiple-value-prog1
			 (values nil (pop instances))
		       (unless instances
			 (setup_becoming (first typesgenerated))))))
		  (class_declass
		   (return-from typeloop (values nil (pop class_declass)))))
	      (unless typesremaining (return t))
	      ;; move first entry from remaining into generated, 
	      ;; this is equivalent to (push (pop typesremaining) typesgenerated)
	      ;; but w/o consing
	      (let ((head typesremaining))
		(setf typesremaining (rest typesremaining))
		(setf (rest head) typesgenerated
		      typesgenerated head))
	      (setf instances (instances (first typesgenerated)))
	      (setup_ceasing (first typesgenerated))
	      (dolist (tp (#+(or lucid allegro)clos:class-direct-subclasses
			   #-(or lucid allegro)class-direct-subclasses
			     (first typesgenerated)))
		(unless (member tp typesgenerated)
		  (pushnew tp typesremaining)))))))
    (clos-enumerable-generator R)))

(defun clos-enumerable-atomic (name keys class-name comclass-name)
  ;; class-name is the name of the COMMITTED class
 (warn-ignored-keys name keys (:generator)
   `(,@(clos-nonenumerable-atomic name keys class-name comclass-name)
     :generator
     (lambda (subsetflg vars rel &rest ignore)
		   (declare (ignore subsetflg vars ignore))
		   `((template (output))
		     (effort ,(* 3 (relationsize '(x) (list rel 'x))))
		     (initialstate clos-enumerable-atomic-generator))))))

(defun nonenumerable-direct (name keys class-name)
  (declare (ignore name keys))
  ;; class-name is the name of the COMMITTED class
  
  ;; for unary relations that represent the direct (no inheritance) committed
  ;; instances of a transition-semantics class, that is not enumerable
  `(:representation individual :arity 1
    :tester (lambda (&rest ignore) (declare (ignore ignore))
	      ;; AP5 requires a tester, but it would be odd to ever
	      ;; actually test the DIRECT instance relation
		    '(lambda (ignore a1) (declare (ignore ignore))
		      (clos-atomic-direct-tester a1 (load-memo (find-class ',class-name)))))
    :testeffort 5
    :adder (lambda (&rest ignore) (declare (ignore ignore))
	      '(lambda (thisrel instance) (declare (ignore thisrel))
		(commit instance)))
    :deleter (lambda (&rest ignore) (declare (ignore ignore))
		     '(lambda (thisrel instance) (declare (ignore thisrel))
		       (decommit instance)))))

(defun enumerable-direct (name keys class-name)
    ;; class-name is the name of the COMMITTED class
  `(,@ (nonenumerable-direct name keys class-name)
       :generator (simplegenerator (output)
		       (instances (load-memo (find-class ',class-name))))))

(defun CLOS-ATOMIC-TRIGGER-SUPER
    (superrel adds dels &aux (tvs (if adds
				      (if dels '(t nil) '(t))
				    '(nil))))
  ;; This trigger is responsible for telling AP5 that SUB(x) leads
  ;; to SUPERREL(x) for every IMMEDIATE clos subrelation of superrel
  ;; (both add and delete)
  (dolist (subclass (#+(or lucid allegro)clos:class-direct-subclasses
		     #-(or lucid allegro)class-direct-subclasses
		     (clos-class-for-relation superrel ?)))
    (when (slot-boundp subclass 'clos-relation)
      (map-updates
       #'(lambda (rel tpl tv) (declare (ignore rel))
		 (if tv
		     (++ funcall superrel (first tpl))
		   (-- funcall superrel (first tpl))))
       :rel (clos-relation subclass)
       :tv tvs))))

(defun update-subclass-triggers (subclassrel new old)
  ;; NEW are the new direct superclass relations of subclassrel
  ;; OLD are the previous direct superclass relations of subclassrel
  ;; we must ADD  TRIGGER entries for each added superclass
  ;; and remove  TRIGGER entries each removed superclass
  ;; the trigger entries look like 
  ;;     (subclassrel t 'CLOS-ATOMIC-TRIGGER-SUPER superclass t)
  ;; and (subclassrel nil 'CLOS-ATOMIC-TRIGGER-SUPER superclass nil)
  (dolist (super new)
    (unless (member super old)
      (++ trigger subclassrel t 'CLOS-ATOMIC-TRIGGER-SUPER super t)
      (++ trigger subclassrel nil 'CLOS-ATOMIC-TRIGGER-SUPER super nil)))
  (dolist (super old)
    (unless (member super new)
      (-- trigger subclassrel t 'CLOS-ATOMIC-TRIGGER-SUPER super t)
      (-- trigger subclassrel nil 'CLOS-ATOMIC-TRIGGER-SUPER super nil)))
  )

(defun committed-class-name (class-name)
	   (intern (format nil "COMMITTED-~a" class-name)
		   (symbol-package class-name)))

(let ((emoclass (find-class 'ec-mixin-object)))
 (defun immediate-clos-superrels (class-name
				&aux (cclass (find-class class-name nil)))
   (when cclass
     (loop with potsup = (decommitted-superclass cclass) and srel
	 for c in (#+(or lucid allegro)clos:class-direct-superclasses
		   #-(or lucid allegro)class-direct-superclasses
		   cclass)
	 unless (or (eq c emoclass) (eq c potsup))
	 when (setf srel (clos-class-for-relation ? c default nil))
	 collect srel)))
)

(defvar *old-sups* nil)

(defmacro define-clos-relation (class-name superclasses slots
				&environment env &rest options
	                        &aux (enump (assoc :enumerable options))
				     (atomicp (assoc :atomic options))
				     (size (assoc :size options)))
  (when size
  ;;If provided, the size attribute value  should be a numeric literal
  ;;estimating the number of TOTAL instances (direct + inherited) of the class.
    (setf options (remove size options)
	  size (second size)))

  ;; USE AN ENUMERABLE IMPLEMENTATION UNLESS options EXPLICITLY INCLUDES
  ;; (:enumerable nil)
  (if enump
      (setf options (remove enump options)
	    enump (second enump))
    (setf enump t))

  ;; USE A TRANSACTION-BASED IMPLEMENTATION UNLESS options EXPLICITLY INCLUDES
  ;; (:atomic nil)
  (if atomicp
      (setf options (remove atomicp options)
	    atomicp (second atomicp))
    (setf atomicp t))
  (unless (or enump atomicp)
    (Error "only use DEFINE-CLOS-RELATION for classes that are either enumerable or have transaction semantics.~%~
           If you want neither, simply use DEFCLASS."))
  (let ((meta (if atomicp 'atomic-potential 'enumerable-nonatomic))
	(comclassname (when atomicp (committed-class-name class-name))))
  `(progn (eval-when (:compile-toplevel :load-toplevel :execute)
	    ,(when atomicp 
	       `(setf *old-sups* (immediate-clos-superrels ',comclassname)))
	    (defclass ,class-name ,superclasses ,slots (:metaclass ,meta)
		,@options))
	  ,.(when atomicp
	      `((eval-when (:compile-toplevel :load-toplevel :execute)
		  (defclass ,comclassname
		      (,class-name
		       ,.(mapcar
			  #'(lambda (super
				     &aux (superc (find-class super t env)))
			      (if (typep superc 'atomic-potential)
				   (committed-class-name super)
				;; a superclass that is NOT a relation
				;; I allow it, but it had better be a pure mixin (virtual) class
				super))
			      superclasses))
		    () (:metaclass ,(if enump
					'enumerable-atomic-committed
				      'nonenumerable-atomic-committed)))
		  (setf (committed-subclass (find-class ',class-name))
		        (find-class ',comclassname)
			(decommitted-superclass (find-class ',comclassname))
			(find-class ',class-name)))
		;; NOW DEFINE THE DIRECT (non derived) RELATION
		(defrelation ,(direct-rel-name class-name)
		   :representation (,(if enump
					 'enumerable-direct
				         'nonenumerable-direct)
				       ,comclassname))
		;; LINK THE POTENTIAL CLASS TO THE DIRECT RELATION
		(eval-when (:compile-toplevel :load-toplevel :execute)
 		  (setf (direct-rel (find-class ',class-name))
		        (theap5relation ,(direct-rel-name class-name))))))
	  ;; NOW DEFINE THE RELATION THAT CORRESPONDS TO THE (committed) CLASS
	  (defrelation ,class-name
	      ,@ (when size `(:size ((output) ,size)))
	      ;;:derivation
	       ,@ (if enump
		    (if atomicp
		      `(:derivation
			(clos-enumerable-atomic ,class-name ,comclassname))
		      `(:representation
			(clos-enumerable-nonatomic ,class-name)))
		   `(:derivation clos-nonenumerable-atomic
				 ,class-name ,comclassname)))
	  ,(when atomicp
	     `(eval-when (:compile-toplevel :load-toplevel :execute)
		(update-subclass-triggers
		 (theap5relation ,class-name)
		 (immediate-clos-superrels ',comclassname)
		 *old-sups*)
		(setf *old-sups* nil)))
	  (theap5relation ,class-name))))
